#include <stdbool.h>
#include <stdio.h>

#include <windows.h>
#include <commctrl.h>
#include <iphlpapi.h>
#include <psapi.h>

#define UNUSED __attribute__((unused))
#define SECTION_SHARED __attribute__((section(".shared"), shared))

#define REGISTRY_KEY "SOFTWARE\\SysTrayUtil"
#define REGISTRY_VALUE_ORDER "Order"

static void wperror(const char *text) {
	fprintf(stderr, "%s: %d.\n", text, (int) GetLastError());
}

static int device_id_format(char *buffer, int size,
	const char *prefix, const char *device_id, const char *const *parts) {
	int index = 0;
	int length = strlen(prefix);
	if (length < size) {
		memcpy(buffer, prefix, length + 1);
		index += length;
	}
	for (int i = 0; parts[i]; i++) {
		const char *substr = strstr(device_id, parts[i]);
		if (!substr || substr == device_id) {
			continue;
		}
		char c = substr[-1];
		if (c != '\\' && c != '&') {
			continue;
		}
		substr = &substr[strlen(parts[i])];
		length = strlen(substr);
		for (int j = 0; j < length; j++) {
			c = substr[j];
			if (c == '\\' || c == '&') {
				length = j;
				break;
			}
		}
		if (index + 1 + length < size) {
			buffer[index++] = '.';
			memcpy(&buffer[index], substr, length);
			index += length;
			buffer[index] = '\0';
		}
	}
	return index;
}

static int tray_id_append_network_hwid(char *buffer, int size,
	const wchar_t *tooltip) {
	int name_length = 0;
	for (int i = 0; tooltip[i]; i++) {
		if (tooltip[i] == L'\n') {
			name_length = i;
			break;
		}
	}
	if (name_length <= 0) {
		name_length = wcslen(tooltip);
	}
	if (name_length >= MAX_PATH) {
		return 0;
	}
	ULONG ssize = 0;
	GetAdaptersInfo(NULL, &ssize);
	if (ssize <= 0) {
		return 0;
	}
	IP_ADAPTER_INFO *infos = malloc(ssize);
	if (!infos) {
		return 0;
	}
	DWORD result = GetAdaptersInfo(infos, &ssize);
	if (result != ERROR_SUCCESS) {
		SetLastError(result);
		wperror("Failed to enumerate network adapters");
		free(infos);
		return 0;
	}
	wchar_t name[name_length + 1];
	char device_id[MAX_PATH] = "\0";
	IP_ADAPTER_INFO *info = infos;
	static const char key_format[] = "SYSTEM\\CurrentControlSet\\"
		"Control\\Network\\{4D36E972-E325-11CE-BFC1-08002BE10318}\\"
		"%s\\Connection";
	int key_format_length = sizeof(key_format) - 2 - 1;
	while (info) {
		char key[key_format_length + strlen(info->AdapterName) + 1];
		sprintf(key, key_format, info->AdapterName);
		info = info->Next;
		HKEY hkey = NULL;
		DWORD status;
		status = RegOpenKeyExA(HKEY_LOCAL_MACHINE, key,
			0, KEY_QUERY_VALUE, &hkey);
		if (status != ERROR_SUCCESS) {
			continue;
		}
		DWORD vsize = 2 * (name_length + 1);
		status = RegQueryValueExW(hkey, L"Name",
			NULL, NULL, (void *) name, &vsize);
		name[name_length] = L'\0';
		if (status != ERROR_SUCCESS) {
			RegCloseKey(hkey);
			continue;
		}
		if (wcsstr(tooltip, name) == tooltip) {
			int length = wcslen(name);
			const wchar_t *c = &tooltip[length];
			if (c[0] != L'\n' && c[0] != L'\0' &&
				!(c[0] == L' ' && c[1] == L'(')) {
				RegCloseKey(hkey);
				continue;
			}
		}
		vsize = sizeof(device_id);
		status = RegQueryValueExA(hkey, "PnpInstanceID",
			NULL, NULL, (void *) device_id, &vsize);
		device_id[sizeof(device_id) - 1] = '\0';
		RegCloseKey(hkey);
		break;
	}
	free(infos);
	if (device_id[0]) {
		if (strstr(device_id, "PCI\\") == device_id) {
			static const char *const parts[] = {
				"VEN_", "DEV_", NULL
			};
			return device_id_format(buffer, size,
				"pci", device_id, parts);
		} else if (strstr(device_id, "USB\\") == device_id) {
			static const char *const parts[] = {
				"VID_", "PID_", "MI_", NULL
			};
			return device_id_format(buffer, size,
				"usb", device_id, parts);
		}
	}
	return 0;
}

static int tray_id_append_systray_main(char *buffer, int size, UINT id) {
	int message = id - WM_USER;
	const char *name = NULL;
	switch (message) {
		case 201: name = "power"; break;
		case 202: name = "hotplug"; break;
		case 203: name = "volume"; break;
		case 204: name = "stickykeys"; break;
		case 205: name = "mousekeys"; break;
		case 206: name = "filterkeys"; break;
	}
	if (name) {
		int length = strlen(name);
		if (length < size) {
			memcpy(buffer, name, length + 1);
			return length;
		}
	} else if (message >= 0) {
		char text[10];
		sprintf(text, "%d", message);
		int length = strlen(text);
		if (length < size) {
			memcpy(buffer, text, length + 1);
			return length;
		}
	}
	return 0;
}

static int tray_id_clear_path(char *buffer, int length) {
	int index = -1;
	for (int i = length - 1; i >= 0; i--) {
		if (buffer[i] == '\\') {
			index = i;
			break;
		}
	}
	if (index < 0) {
		return length;
	}
	for (int i = length; i >= index + 1; i--) {
		buffer[i - index - 1] = buffer[i];
	}
	length = length - index - 1;
	if (length >= 4 && buffer[length - 4] == '.' &&
		(buffer[length - 3] == 'e' || buffer[length - 3] == 'E') &&
		(buffer[length - 2] == 'x' || buffer[length - 2] == 'X') &&
		(buffer[length - 1] == 'e' || buffer[length - 1] == 'E')) {
		length -= 4;
		buffer[length] = '\0';
	}
	return length;
}

static int tray_id_simplify(char *buffer, int length, bool lower) {
	for (int i = 0; i < length; i++) {
		char c = buffer[i];
		if (!(c >= 'A' && c <= 'Z') &&
			!(c >= 'a' && c <= 'z') &&
			!(c >= '0' && c <= '9')) {
			for (int j = i; j < length; j++) {
				buffer[j] = buffer[j + 1];
			}
			length--;
		} else if (lower && c >= 'A' && c <= 'Z') {
			buffer[i] = c - 'A' + 'a';
		}
	}
	return length;
}

static char *tray_id_from_button(const TBBUTTON *button, bool sorting) {
	if (!button || (button->fsState & TBSTATE_HIDDEN) || !button->dwData) {
		return NULL;
	}
	struct {
		HWND window;
		UINT id;
		UINT message;
	} *tray = (void *) button->dwData;
	if (!tray->window) {
		return NULL;
	}
	DWORD pid = 0;
	if (!GetWindowThreadProcessId(tray->window, &pid) || !pid) {
		wperror("Failed to get pid");
		return NULL;
	}
	HANDLE process = OpenProcess(PROCESS_QUERY_INFORMATION, FALSE, pid);
	if (!process) {
		wperror("Failed to open process");
		return NULL;
	}
	char buffer[2 * MAX_PATH];
	if (GetProcessImageFileNameA(process, buffer, sizeof(buffer)) <= 0) {
		wperror("Failed to get process file name");
		CloseHandle(process);
		return NULL;
	}
	CloseHandle(process);
	int length = strlen(buffer);
	length = tray_id_clear_path(buffer, length);
	length = tray_id_simplify(buffer, length, true);
	if (length <= 0) {
		return NULL;
	}
	int divider = -1;
	if (length + 1 < (int) sizeof(buffer)) {
		divider = length++;
		buffer[divider] = ':';
		buffer[divider + 1] = '\0';
		if (length + 1 < (int) sizeof(buffer)) {
			int old_length = length;
			GetClassNameA(tray->window, &buffer[length],
				sizeof(buffer) - length);
			length = strlen(buffer);
			length = tray_id_simplify(&buffer[divider + 1],
				length - divider - 1, false) + divider + 1;
			if (length == old_length) {
				buffer[divider] = '\0';
				divider = -1;
				length--;
			}
		}
	}
	if (length <= 0) {
		return NULL;
	}
	int extra = 0;
	char *extra_buffer = &buffer[length + 1];
	int extra_size = sizeof(buffer) - length - 1;
	if (!strcmp(buffer, "explorer:ConnectionsTray") &&
		button->iString && !IS_INTRESOURCE(button->iString)) {
		if (button->iString == -1) {
			if (sorting) {
				SendMessage(tray->window, tray->message,
					tray->id, WM_MOUSEMOVE);
			}
		} else {
			extra = tray_id_append_network_hwid(extra_buffer,
				extra_size, (const wchar_t *) button->iString);
		}
	} else if (!strcmp(buffer, "explorer:SystemTrayMain") &&
		tray->id == tray->message) {
		extra = tray_id_append_systray_main(extra_buffer,
			extra_size, tray->id);
	}
	if (extra > 0) {
		buffer[length] = ':';
		length += extra + 1;
	}
	char *result = malloc(length + 1);
	if (!result) {
		return NULL;
	}
	memcpy(result, buffer, length + 1);
	return result;
}

static char *tray_id_from_index(HWND tray_window, int index, bool sorting) {
	TBBUTTON button;
	if (!SendMessage(tray_window,
		TB_GETBUTTON, index, (LPARAM) &button)) {
		fprintf(stderr, "Failed to get button.\n");
		return NULL;
	}
	return tray_id_from_button(&button, sorting);
}

static char **sorted = NULL;

static void sorted_free() {
	for (int i = 0; sorted && sorted[i]; i++) {
		free(sorted[i]);
	}
	free(sorted);
	sorted = NULL;
}

static void sorted_load() {
	sorted_free();
	HKEY hkey;
	DWORD status;
	status = RegOpenKeyExA(HKEY_CURRENT_USER, REGISTRY_KEY,
		0, KEY_QUERY_VALUE, &hkey);
	if (status != ERROR_SUCCESS) {
		SetLastError(status);
		wperror("Failed to open registry key");
		return;
	}
	DWORD size;
	status = RegQueryValueExA(hkey, REGISTRY_VALUE_ORDER,
		NULL, NULL, NULL, &size);
	if (status != ERROR_SUCCESS) {
		SetLastError(status);
		wperror("Failed to query registry value");
		RegCloseKey(hkey);
		return;
	}
	char *data = malloc(size);
	if (!data) {
		fprintf(stderr, "Failed to allocate memory.\n");
		RegCloseKey(hkey);
		return;
	}
	DWORD type;
	status = RegQueryValueExA(hkey, REGISTRY_VALUE_ORDER,
		NULL, &type, (void *) data, &size);
	if (status != ERROR_SUCCESS) {
		SetLastError(status);
		wperror("Failed to query registry value");
		RegCloseKey(hkey);
		free(data);
		return;
	}
	if (type != REG_MULTI_SZ) {
		fprintf(stderr, "Illegal value type.\n");
		RegCloseKey(hkey);
		free(data);
		return;
	}
	int start = 0;
	int index = 0;
	for (int i = 0; i < (int) size; i++) {
		if (data[i] == '\0') {
			if (strlen(&data[start]) > 0) {
				index++;
			}
			start = i + 1;
		}
	}
	start = 0;
	if (index > 0) {
		sorted = malloc(sizeof(*sorted) * (index + 1));
	}
	if (sorted) {
		memset(sorted, 0, sizeof(*sorted) * (index + 1));
		bool success = true;
		index = 0;
		for (int i = 0; i < (int) size; i++) {
			if (data[i] == '\0') {
				if (strlen(&data[start]) > 0) {
					char *id = strdup(&data[start]);
					if (!id) {
						success = false;
						break;
					}
					sorted[index++] = id;
				}
				start = i + 1;
			}
		}
		if (!success) {
			sorted_free();
		}
	}
	RegCloseKey(hkey);
	free(data);
}

static void sorted_save() {
	int length = 0;
	for (int i = 0; sorted && sorted[i]; i++) {
		length += strlen(sorted[i]) + 1;
	}
	length++;
	char *data = malloc(length);
	if (!data) {
		return;
	}
	int index = 0;
	for (int i = 0; sorted && sorted[i]; i++) {
		int ilength = strlen(sorted[i]) + 1;
		memcpy(&data[index], sorted[i], ilength);
		index += ilength;
	}
	data[index++] = '\0';
	HKEY hkey;
	int status;
	status = RegCreateKeyExA(HKEY_CURRENT_USER, REGISTRY_KEY,
		0, NULL, 0, KEY_ALL_ACCESS, NULL, &hkey, NULL);
	if (status != ERROR_SUCCESS) {
		SetLastError(status);
		wperror("Failed to open registry key");
		free(data);
		return;
	}
	status = RegSetValueExA(hkey, REGISTRY_VALUE_ORDER,
		0, REG_MULTI_SZ, (void *) data, length);
	if (status != ERROR_SUCCESS) {
		SetLastError(status);
		wperror("Failed to set registry value");
	}
	RegCloseKey(hkey);
	free(data);
}

static int sorted_index_of(const char *id) {
	if (!id) {
		return -1;
	}
	for (int i = 0; sorted && sorted[i]; i++) {
		if (!strcmp(sorted[i], id)) {
			return i;
		}
	}
	return -1;
}

struct tray_item_t {
	int index;
	char *id;
};

static bool sorting = false;

static int tray_sort_compare(const void *a, const void *b) {
	const struct tray_item_t *ta = a;
	const struct tray_item_t *tb = b;
	int ia = sorted_index_of(ta->id);
	int ib = sorted_index_of(tb->id);
	if (ia < 0) {
		return -1;
	} else if (ib < 0) {
		return 1;
	} else if (ia < ib) {
		return -1;
	} else if (ia > ib) {
		return 1;
	} else {
		return 0;
	}
}

static void tray_sort(HWND tray_window, bool print) {
	int count = (int) SendMessage(tray_window, TB_BUTTONCOUNT, 0, 0);
	if (count <= 0) {
		return;
	}
	struct tray_item_t items[count];
	for (int i = 0; i < count; i++) {
		struct tray_item_t *item = &items[i];
		item->index = i;
		item->id = tray_id_from_index(tray_window, i, true);
	}
	qsort(items, count, sizeof(*items), tray_sort_compare);
	sorting = true;
	for (int i = 0; i < count; i++) {
		struct tray_item_t *item = &items[i];
		if (item->index != i) {
			int from = item->index;
			SendMessage(tray_window,
				TB_MOVEBUTTON, item->index, i);
			for (int j = i + 1; j < count; j++) {
				item = &items[j];
				if (item->index < from) {
					item->index++;
				}
			}
		}
	}
	sorting = false;
	for (int i = 0; i < count; i++) {
		char *id = items[i].id;
		if (print) {
			printf("Set index %d for \"%s\".\n",
				i, id ? id : "null");
		}
		free(id);
	}
}

static bool sorted_inc_size(int current) {
	char **data = sorted
		? realloc(sorted, sizeof(*sorted) * (current + 2))
		: malloc(sizeof(*sorted) * 2);
	if (!data) {
		return false;
	}
	if (!sorted) {
		data[0] = NULL;
	}
	sorted = data;
	return true;
}

static void sorted_set_before(const char *what, const char *before) {
	if (!what) {
		return;
	}
	int what_index = sorted_index_of(what);
	int before_index = sorted_index_of(before);
	if (what_index >= 0 && (what_index == before_index - 1 ||
		what_index == before_index)) {
		return;
	}
	int count;
	for (count = 0; sorted && sorted[count]; count++);
	if (!before) {
		before_index = count;
	} else if (before_index < 0) {
		char *before_copy = strdup(before);
		if (!before_copy || !sorted_inc_size(count++)) {
			free(before_copy);
			return;
		}
		for (int i = count; i > 0; i--) {
			sorted[i] = sorted[i - 1];
		}
		sorted[0] = before_copy;
		before_index = 0;
		if (what_index >= 0) {
			what_index++;
		}
	}
	if (what_index >= 0) {
		char *what_copy = sorted[what_index];
		if (what_index >= before_index) {
			for (int i = what_index; i > before_index; i--) {
				sorted[i] = sorted[i - 1];
			}
			sorted[before_index] = what_copy;
		} else {
			for (int i = what_index; i < before_index - 1; i++) {
				sorted[i] = sorted[i + 1];
			}
			sorted[before_index - 1] = what_copy;
		}
	} else {
		char *what_copy = strdup(what);
		if (!what_copy || !sorted_inc_size(count++)) {
			free(what_copy);
			return;
		}
		for (int i = count; i > before_index; i--) {
			sorted[i] = sorted[i - 1];
		}
		sorted[before_index] = what_copy;
	}
}

static void tray_move(HWND tray_window, int from, int to) {
	if (from == to) {
		return;
	}
	int count = (int) SendMessage(tray_window, TB_BUTTONCOUNT, 0, 0);
	if (count <= 0) {
		return;
	}
	char *what = tray_id_from_index(tray_window, from, false);
	if (!what) {
		return;
	}
	char *next = NULL;
	for (int i = count - 1; i >= 0; i--) {
		char *id = i == from ? what :
			tray_id_from_index(tray_window, i, false);
		if (!id) {
			continue;
		}
		if (sorted_index_of(id) < 0) {
			sorted_set_before(id, next);
		}
		if (next != what) {
			free(next);
		}
		next = id;
	}
	if (next != what) {
		free(next);
	}
	char *before = NULL;
	int before_index = to >= from ? to + 1 : to;
	while (!before && before_index < count) {
		before = tray_id_from_index(tray_window,
			before_index++, false);
	}
	sorted_set_before(what, before);
	free(before);
	free(what);
	sorted_save();
}

static HWND get_tray_window() {
	HWND window = FindWindowA("Shell_TrayWnd", NULL);
	if (window) {
		window = FindWindowExA(window, NULL, "TrayNotifyWnd", NULL);
	}
	if (window) {
		window = FindWindowExA(window, NULL, "SysPager", NULL);
	}
	if (window) {
		window = FindWindowExA(window, NULL, "ToolbarWindow32", NULL);
	}
	return window;
}

static void tray_print(HWND tray_window, const char *message,
	const TBBUTTON *buttons, int index_count) {
	int count = buttons ? index_count : 1;
	for (int i = 0; i < count; i++) {
		char *id = buttons ? tray_id_from_button(&buttons[i], false)
			: tray_id_from_index(tray_window, index_count, false);
		printf("%s \"%s\".\n", message, id ? id : "null");
		free(id);
	}
}

enum {
	XX_TRAY_SORT = (WM_USER << 1) + 1
};

enum process_type_t {
	PROCESS_TYPE_INVALID,
	PROCESS_TYPE_MAIN,
	PROCESS_TYPE_EXPLORER
};

static enum process_type_t process_type = PROCESS_TYPE_INVALID;
static HWND tray_window = NULL;
static bool has_console = false;

static LRESULT CALLBACK hook_msg(int code, WPARAM wparam, LPARAM lparam) {
	if (process_type == PROCESS_TYPE_EXPLORER &&
		code == HC_ACTION && wparam == PM_REMOVE) {
		MSG *msg = (void *) lparam;
		bool tray_sort_event = msg->message == XX_TRAY_SORT;
		if (tray_sort_event && msg->hwnd == tray_window) {
			MSG check_msg;
			if (!PeekMessageA(&check_msg, tray_window,
				XX_TRAY_SORT, XX_TRAY_SORT, PM_NOREMOVE)) {
				printf("Sort tray.\n");
				tray_sort(tray_window, has_console);
			}
		}
	}
	return CallNextHookEx(NULL, code, wparam, lparam);
}

static LRESULT CALLBACK hook_call(int code, WPARAM wparam, LPARAM lparam) {
	if (process_type == PROCESS_TYPE_EXPLORER && code == HC_ACTION) {
		CWPSTRUCT *call = (void *) lparam;
		bool add_buttons_event = call->message == TB_ADDBUTTONSA ||
			call->message == TB_ADDBUTTONSW;
		bool insert_button_event = call->message == TB_INSERTBUTTONA ||
			call->message == TB_INSERTBUTTONW;
		bool move_button_event = call->message == TB_MOVEBUTTON &&
			!sorting;
		static bool sort_initial = true;
		bool sort_tray = sort_initial;
		if ((add_buttons_event || insert_button_event ||
			move_button_event) && call->hwnd == tray_window) {
			sort_tray = true;
			if (add_buttons_event && has_console) {
				const TBBUTTON *buttons = (void *) call->lParam;
				tray_print(tray_window, "Add button",
					buttons, call->wParam);
			}
			if (insert_button_event && has_console) {
				const TBBUTTON *button = (void *) call->lParam;
				tray_print(tray_window, "Insert button",
					button, 1);
			}
			if (move_button_event) {
				if (has_console) {
					tray_print(tray_window, "Move button",
						NULL, call->wParam);
				}
				tray_move(tray_window,
					call->wParam, call->lParam);
			}
		}
		if (sort_tray) {
			sort_initial = false;
			PostMessage(tray_window, XX_TRAY_SORT, 0, 0);
		}
	}
	return CallNextHookEx(NULL, code, wparam, lparam);
}

SECTION_SHARED static DWORD main_pid = 0;
SECTION_SHARED static bool attach_console = false;

static BOOL WINAPI console_handler_ignore(UNUSED DWORD ctrl_type) {
	return ctrl_type == CTRL_C_EVENT || ctrl_type == CTRL_CLOSE_EVENT;
}

BOOL WINAPI DllMain(UNUSED HINSTANCE instance,
	DWORD reason, UNUSED LPVOID reserved) {
	if (reason == DLL_PROCESS_ATTACH) {
		DWORD pid = GetCurrentProcessId();
		tray_window = get_tray_window();
		if (tray_window) {
			DWORD tray_pid;
			if (GetWindowThreadProcessId(tray_window,
				&tray_pid) && tray_pid == pid) {
				process_type = PROCESS_TYPE_EXPLORER;
			}
		}
		if (process_type != PROCESS_TYPE_EXPLORER && !main_pid) {
			main_pid = pid;
			process_type = PROCESS_TYPE_MAIN;
		}
	}
	if (process_type != PROCESS_TYPE_EXPLORER) {
		return process_type == PROCESS_TYPE_MAIN;
	}
	if (reason == DLL_PROCESS_ATTACH) {
		has_console = attach_console && AttachConsole(main_pid);
		if (has_console) {
			freopen("CONOUT$", "w", stdout);
			freopen("CONOUT$", "w", stderr);
			SetConsoleCtrlHandler(console_handler_ignore, TRUE);
		}
		printf("Attach DLL.\n");
		sorted_load();
	} else if (reason == DLL_PROCESS_DETACH) {
		printf("Detach DLL.\n");
		sorted_free();
		if (has_console) {
			SetConsoleCtrlHandler(console_handler_ignore, FALSE);
			FreeConsole();
		}
	}
	return TRUE;
}

static bool check_sleep(int ms) {
	MSG msg;
	bool value;
	DWORD error = ERROR_SUCCESS;
	UINT_PTR timer = SetTimer(NULL, 0, ms, NULL);
	while (true) {
		int result = GetMessageA(&msg, NULL, 0, 0);
		if (result == -1) {
			value = false;
			error = GetLastError();
			break;
		} else if (result == 0) {
			value = false;
			break;
		} else if (msg.message == WM_TIMER &&
			(UINT_PTR) msg.wParam == timer) {
			value = true;
			break;
		}
	}
	KillTimer(NULL, timer);
	if (!value) {
		SetLastError(error);
		wperror("Finish sleep");
	}
	return value;
}

static void inject(bool debug) {
	if (debug) {
		if (AllocConsole()) {
			attach_console = true;
			freopen("CONOUT$", "w", stdout);
			freopen("CONOUT$", "w", stderr);
		} else {
			MessageBoxA(NULL, "Failed to allocate console.",
				NULL, MB_ICONERROR);
			return;
		}
	}
	HMODULE module;
	HOOKPROC hook_msg_fptr = hook_msg;
	HOOKPROC hook_call_fptr = hook_call;
	void *func_ptr = (void *) &hook_call_fptr;
	func_ptr = *(void **) func_ptr;
	if (!GetModuleHandleEx(GET_MODULE_HANDLE_EX_FLAG_FROM_ADDRESS |
		GET_MODULE_HANDLE_EX_FLAG_UNCHANGED_REFCOUNT,
		func_ptr, &module)) {
		wperror("Failed to get module handle");
		if (debug) {
			check_sleep(2000);
		}
		return;
	}
	SetProcessShutdownParameters(0, 0);
	while (true) {
		printf("Waiting for tray window.\n");
		HWND tray_window = NULL;
		bool finish = false;
		while (true) {
			tray_window = get_tray_window();
			if (tray_window) {
				break;
			} else if (!check_sleep(1000)) {
				finish = true;
				break;
			}
		}
		if (finish) {
			break;
		}
		DWORD tid = GetWindowThreadProcessId(tray_window, NULL);
		HANDLE thread = tid ? OpenThread(SYNCHRONIZE,
			FALSE, tid) : NULL;
		if (thread) {
			HHOOK handle_msg = SetWindowsHookEx(WH_GETMESSAGE,
				hook_msg_fptr, module, tid);
			HHOOK handle_call = SetWindowsHookEx(WH_CALLWNDPROC,
				hook_call_fptr, module, tid);
			if (handle_msg && handle_call) {
				printf("Waiting for thread finish.\n");
				WaitForSingleObject(thread, INFINITE);
				UnhookWindowsHookEx(handle_msg);
				UnhookWindowsHookEx(handle_call);
				CloseHandle(thread);
			} else {
				if (handle_msg) {
					UnhookWindowsHookEx(handle_msg);
				}
				if (handle_call) {
					UnhookWindowsHookEx(handle_call);
				}
				wperror("Failed to set hook");
				CloseHandle(thread);
				if (!check_sleep(1000)) {
					break;
				}
			}
		} else {
			wperror("Failed to open thread");
			if (!check_sleep(1000)) {
				break;
			}
		}
	}
}

__declspec(dllexport) void run() {
	inject(false);
}

__declspec(dllexport) void debug() {
	inject(true);
}
